@@ -1,7 +1,8 @@
 DEFAULT_PORT = 8563
 DEFAULT_AUTOCOMMIT = True
 
-DEFAULT_SOCKET_TIMEOUT = 10
+DEFAULT_CONNECTION_TIMEOUT = 10
+DEFAULT_SOCKET_TIMEOUT = 30
 DEFAULT_QUERY_TIMEOUT = 0
 
 DEFAULT_FETCHMANY_SIZE = 10000
@@ -10,4 +11,6 @@ DEFAULT_FETCH_SIZE_BYTES = 5 * 1024 * 1024
 DRIVER_NAME = 'PyEXASOL'
 
 LOGGER_FILENAME_TIMESTAMP_FORMAT = '%Y%m%d_%H%M%S_%f'
-LOGGER_MAX_JSON_LENGTH = 50000
+LOGGER_MAX_JSON_LENGTH = 20000
+
+EXCEPTION_QUERY_TEXT_MAX_LENGTH = 20000
